@@ -40,7 +40,7 @@
 
    -----
 
-   canon_pp.c: $Revision: 1.10 $
+   canon_pp.c: $Revision: 1.11 $
 
    This file is part of the canon_pp backend, supporting Canon FBX30P 
    and NX40P scanners
@@ -50,7 +50,7 @@
 #include  <lalloca.h>		/* MUST come first for AIX! */
 #endif
 
-#define VERSION "$Revision: 1.10 $"
+#define VERSION "$Revision: 1.11 $"
 #define BACKEND_NAME canon_pp
 
 #define THREE_BITS 0xE0
@@ -1465,6 +1465,9 @@ sane_exit (void)
 	}
 
 	first_dev = NULL;
+	def_scanner = NULL;
+	read_leftover = NULL;
+	num_devices = 0;
 
 	/* FIXEDME: this created a segfault in DLL code. */
 	/* Bug was fixed in libieee1284 0.1.5 */
