@@ -29,10 +29,10 @@ enum fujitsu_Option
   OPT_Y_RES,
 
   OPT_GEOMETRY_GROUP,
-  OPT_TL_X,			/* in mm/2^16 */
-  OPT_TL_Y,			/* in mm/2^16 */
-  OPT_BR_X,			/* in mm/2^16 */
-  OPT_BR_Y,			/* in mm/2^16 */
+  OPT_TL_X,                     /* in mm/2^16 */
+  OPT_TL_Y,                     /* in mm/2^16 */
+  OPT_BR_X,                     /* in mm/2^16 */
+  OPT_BR_Y,                     /* in mm/2^16 */
   OPT_PAGE_WIDTH,
   OPT_PAGE_HEIGHT,
 
@@ -79,63 +79,117 @@ enum fujitsu_Option
   OPT_GREEN_OFFSET,
   OPT_USE_SWAPFILE,
 
+  OPT_IMPRINTER_GROUP,
+  OPT_IMPRINTER,
+  OPT_IMPRINTER_DIR,
+  OPT_IMPRINTER_YOFFSET,
+  OPT_IMPRINTER_STRING,
+  OPT_IMPRINTER_CTR_INIT,
+  OPT_IMPRINTER_CTR_STEP,
+  OPT_IMPRINTER_CTR_DIR,
+
+  OPT_SLEEP_MODE,
+
   /* must come last: */
   NUM_OPTIONS
 };
 
 
+typedef enum {                          /* hardware connection to the scanner */
+        SANE_FUJITSU_NODEV,             /* default, no HW specified yet */
+        SANE_FUJITSU_SCSI,              /* SCSI interface */
+        SANE_FUJITSU_PIO,               /* parallel interface */
+        SANE_FUJITSU_USB                /* USB interface */
+} Fujitsu_Connection_Type;
+
 struct fujitsu
 {
   struct fujitsu *next;
 
   SANE_Option_Descriptor opt[NUM_OPTIONS];
+  Option_Value val[NUM_OPTIONS];
   SANE_Device sane;
 
+
   /* Immutable values which are set during initial probing of the scanner. */
   /* --------------------------------------------------------------------- */
 
-  char vendorName[9];		/* raw data as returned by SCSI inquiry.     */
-  char productName[17];		/* raw data as returned by SCSI inquiry.     */
-  char versionName[5];		/* raw data as returned by SCSI inquiry.     */
+  char vendorName[9];           /* raw data as returned by SCSI inquiry.     */
+  char productName[17];         /* raw data as returned by SCSI inquiry.     */
+  char versionName[5];          /* raw data as returned by SCSI inquiry.     */
 
-  int model;			/* The scanner model.                        */
+  int model;                    /* The scanner model.                        */
 
-  char *devicename;		/* The name of the scanner device.           */
-  int sfd;			/* The scanner device file descriptor.       */
+  char *devicename;             /* The name of the scanner device.           */
 
-  int color_raster_offset;	/* offset between r and b scan line and    */
+  Fujitsu_Connection_Type connection;   /* hardware interface type */
+
+  int sfd;                      /* The scanner device file descriptor.       */
+
+  int color_raster_offset;      /* offset between r and b scan line and    */
   /* between b and g scan line (4 for all      */
   /* known models)                             */
-  int duplex_raster_offset;	/* offset between front and rear page when   */
+  int duplex_raster_offset;     /* offset between front and rear page when   */
   /* when scanning duplex                   */
 
-  int can_read_alternate;	/* duplex transfer mode front/back/front/back... */
-  int adf_present;		/* true if an ADF has been detected.       */
-  int duplex_present;		/* true if a duplex unit has been detected. */
-  int ipc_present;		/* true if ipc2/3 option detected         */
-  int cmp_present;		/* true if cmp2 present                   */
-  int has_hw_status;		/* true for M409X series                  */
-  int has_outline;		/* ............                           */
-  int has_emphasis;		/* ............                           */
-  int has_autosep;		/* ............                           */
-  int has_mirroring;		/* ............                           */
-  int has_white_level_follow;	/* ............                           */
-  int has_subwindow;		/* ............                            */
-  int has_dropout_color;	/*  */
+  int can_read_alternate;       /* duplex transfer mode front/back/front/back... */
+  int read_mode;                /* constants READ_MODE_* and funcs convert_* */
+  int has_adf;                  /* true if an ADF has been detected.       */
+  int has_fb;
+  int duplex_present;           /* true if a duplex unit has been detected. */
+  int ipc_present;              /* true if ipc2/3 option detected         */
+  int cmp_present;              /* true if cmp2 present                   */
+  int has_hw_status;            /* true for M409X series                  */
+  int has_outline;              /* ............                           */
+  int has_emphasis;             /* ............                           */
+  int has_autosep;              /* ............                           */
+  int has_mirroring;            /* ............                           */
+  int has_reverse;
+  int has_white_level_follow;   /* ............                           */
+  int has_subwindow;            /* ............                            */
+  int has_dropout_color;        /*  */
+  int has_imprinter;
+  int has_threshold;
+  int has_brightness;
+  int has_contrast;
+  int has_gamma;
+
+  SANE_Range adf_width_range;
+  SANE_Range adf_height_range;
+  SANE_Range x_range;
+  SANE_Range y_range;
+
+  SANE_Range x_res_range;
+  SANE_Range y_res_range;
+  SANE_Range x_grey_res_range;
+  SANE_Range y_grey_res_range;
+
+  SANE_String_Const vpd_mode_list[7];
+  SANE_String_Const compression_mode_list[9];
+
+  SANE_Int x_res_list[17];      /* range of x and y resolution */
+  SANE_Int y_res_list[17];      /* if scanner has only fixed resolutions */
+  SANE_Int x_res_list_grey[17];
+  SANE_Int y_res_list_grey[17];
+
+  SANE_Range threshold_range;
+  SANE_Range brightness_range;
+  SANE_Range contrast_range;
+
   /* User settable values (usually mirrored in SANE options)               */
   /* --------------------------------------------------------------------- */
 
-  int duplex_mode;		/* Use DUPLEX_* constants below              */
-  int resolution_x;		/* X resolution in dpi                       */
-  int resolution_y;		/* Y resolution in dpi                       */
-  int resolution_linked;	/* When true, Y resolution is set whenever   */
+  int duplex_mode;              /* Use DUPLEX_* constants below              */
+  int resolution_x;             /* X resolution in dpi                       */
+  int resolution_y;             /* Y resolution in dpi                       */
+  int resolution_linked;        /* When true, Y resolution is set whenever   */
   /* X resolution gets changed. true initially, */
   /* becomes false if Y resolution is ever     */
   /* set independently                         */
-  int top_left_x;		/* The desired size of the scan, in          */
-  int top_left_y;		/*   mm/2^16. Notice that there's a second   */
-  int bottom_right_x;		/*   group of these with the "rounded"       */
-  int bottom_right_y;		/*   prefix, below.                          */
+  int top_left_x;               /* The desired size of the scan, in          */
+  int top_left_y;               /*   mm/2^16. Notice that there's a second   */
+  int bottom_right_x;           /*   group of these with the "rounded"       */
+  int bottom_right_y;           /*   prefix, below.                          */
 
   /* A note on the scan are size (topLeft/bottomRight) values. When the    
    * user (the front-end) sets one of these, we change the above values   
@@ -163,11 +217,11 @@ struct fujitsu
    * function.
    */
 
-  int page_width;		/* paper width and height may influence the  */
-  int page_height;		/* way the ADF operates  - in 1/1200 inch    */
+  int page_width;               /* paper width and height may influence the  */
+  int page_height;              /* way the ADF operates  - in 1/1200 inch    */
 
-  int output_depth;		/* How many bits per pixel the user wants.   */
-  int scanner_depth;		/* How many bpp the scanner will send.       */
+  int output_depth;             /* How many bits per pixel the user wants.   */
+  int scanner_depth;            /* How many bpp the scanner will send.       */
 
   /* A note on "depth" values. Many colour scanners require you to use
    * the value "8" for depth if you're doing colour scans even if they 
@@ -183,59 +237,42 @@ struct fujitsu
    * "on the fly", and output_depth would be 4, while scanner_depth is 8.
    */
 
-  int color_mode;		/* Color/Gray/..., use MODE_* constants      */
-  int use_adf;			/* Whether the ADF should be used or not.    */
-  int use_temp_file;		/* Whether to use a temp file for duplex.    */
-  int mirror;			/* Whether to mirror scan area for rear side */
-  int lamp_color;		/* lamp color to use for monochrome scans    */
-  int green_offset;		/* color tuning - green scan line offset     */
-  int blue_offset;		/* color tuning - blue scan line offset      */
+  int color_mode;               /* Color/Gray/..., use MODE_* constants      */
+  int use_adf;                  /* Whether the ADF should be used or not.    */
+  int use_temp_file;            /* Whether to use a temp file for duplex.    */
+  int mirror;                   /* Whether to mirror scan area for rear side */
+  int lamp_color;               /* lamp color to use for monochrome scans    */
+  int green_offset;             /* color tuning - green scan line offset     */
+  int blue_offset;              /* color tuning - blue scan line offset      */
 
-  SANE_Range adf_width_range;
-  SANE_Range adf_height_range;
-  SANE_Range x_range;
-  SANE_Range y_range;
-
-  SANE_Range x_res_range;
-  SANE_Range y_res_range;
-  SANE_Range x_grey_res_range;
-  SANE_Range y_grey_res_range;
-
-  SANE_String_Const vpd_mode_list[7];
-  SANE_String_Const compression_mode_list[9];
-
-  SANE_Int x_res_list[17];	/* range of x and y resolution */
-  SANE_Int y_res_list[17];	/* if scanner has only fixed resolutions */
-  SANE_Int x_res_list_grey[17];
-  SANE_Int y_res_list_grey[17];
 
 
   /* Derived values (calculated by calculateDerivedValues from the above)  */
   /* --------------------------------------------------------------------- */
 
-  int rounded_top_left_x;	/* Same as topLeft... and bottomRight...,    */
-  int rounded_top_left_y;	/*   but "rounded" to the nearest 1/1200     */
-  int rounded_bottom_right_x;	/*   value, i.e. these are the values really */
-  int rounded_bottom_right_y;	/*   used for scanning.                      */
-  int top_margin;		/* Top margin in 1/1200 inch                 */
-  int left_margin;		/* Left margin in 1/1200 inch                */
-  int scan_width;		/* Width of scan in 1/1200 inch              */
-  int scan_height;		/* Height of scan in 1/1200 inch             */
-  int scan_width_pixels;	/* Pixels per scan line for this job         */
-  int scan_height_pixels;	/* Number of lines in this job               */
-  int bytes_per_scan_line;	/* Number of bytes per line in this job      */
+  int rounded_top_left_x;       /* Same as topLeft... and bottomRight...,    */
+  int rounded_top_left_y;       /*   but "rounded" to the nearest 1/1200     */
+  int rounded_bottom_right_x;   /*   value, i.e. these are the values really */
+  int rounded_bottom_right_y;   /*   used for scanning.                      */
+  int top_margin;               /* Top margin in 1/1200 inch                 */
+  int left_margin;              /* Left margin in 1/1200 inch                */
+  int scan_width;               /* Width of scan in 1/1200 inch              */
+  int scan_height;              /* Height of scan in 1/1200 inch             */
+  int scan_width_pixels;        /* Pixels per scan line for this job         */
+  int scan_height_pixels;       /* Number of lines in this job               */
+  int bytes_per_scan_line;      /* Number of bytes per line in this job      */
 
   /* Operational values (contain internal status information etc.)         */
   /* --------------------------------------------------------------------- */
 
-  int default_pipe;		/* Pipe between reader process and backend.  */
-  int duplex_pipe;		/* Additional pipe for duplex scans.         */
-  int reader_pid;		/* Process ID of the reader process.         */
+  int default_pipe;             /* Pipe between reader process and backend.  */
+  int duplex_pipe;              /* Additional pipe for duplex scans.         */
+  int reader_pid;               /* Process ID of the reader process.         */
 
-  int reverse;			/* Whether to reverse the image. Not user    */
+  int reverse;                  /* Whether to reverse the image. Not user    */
   /* settable but required internally.         */
 
-  int i_transfer_length;	/* needed when the scanner returns           */
+  int i_transfer_length;        /* needed when the scanner returns           */
   /* compressed data and size of return data   */
   /* is unknown in advance */
 
@@ -254,8 +291,8 @@ struct fujitsu
    */
   int eof;
 
-  unsigned char *buffer;	/* Buffer used for SCSI transfers.           */
-  unsigned int scsi_buf_size;	/* Max amount of data in one SCSI call.      */
+  unsigned char *buffer;        /* Buffer used for SCSI transfers.           */
+  unsigned int scsi_buf_size;   /* Max amount of data in one SCSI call.      */
 
   /** these can be set */
 
@@ -263,7 +300,6 @@ struct fujitsu
   int threshold;
   int contrast;
 
-  int rif;
   int bitorder;
   int compress_type;
   int compress_arg;
@@ -294,6 +330,16 @@ struct fujitsu
   /***** end of "set window" terms *****/
 
   int dropout_color;
+
+  SANE_Bool use_imprinter;
+  int imprinter_direction;
+  SANE_Word imprinter_y_offset;
+  SANE_Char imprinter_string[max_imprinter_string_length];
+  int imprinter_ctr_init;
+  int imprinter_ctr_step;
+  int imprinter_ctr_dir;
+
+  SANE_Int sleep_time;
 };
 
 /** scan only front page */
@@ -308,14 +354,17 @@ struct fujitsu
 #define MODEL_3091 1
 /** Fujitsu M3096 (400x400dpi grayscale, simplex, FB and optional ADF) */
 #define MODEL_3096 2
-/** Fujitsu ScanPartner 15C (....) */
-#define MODEL_SP15 3
 /** 3093 */
 #define MODEL_3093 4
 /** 4097 */
 #define MODEL_4097 5
 /** fi-???? */
 #define MODEL_FI   6
+#define MODEL_3097 7
+/** Fujitsu M3092DCd (300x600dpi, color, duplex, FB and ADF) */
+#define MODEL_3092 8
+/** ScanPartner fi-4120 and ScanParter fi-4220 scsi/usb */
+#define MODEL_FI4x20   9
 
 /* A note regarding the MODEL... constants. There's a place in
  * identifyScanner() where the INQUIRY data is parsed and the model
@@ -328,7 +377,7 @@ struct fujitsu
  * treated differently in the backend. For example, if the "abc" model
  * has an optional ADF and the "xyz" model has the ADF built-in, we
  * don't need to distinguish these models, and it's sufficient to set
- * the "adf_present" variable accordingly.
+ * the "has_adf" variable accordingly.
  * 
  * The same would apply if two sub-models behave in the same way but
  * one of them is faster, or has some exotic extra capability that we
@@ -364,6 +413,11 @@ struct fujitsu
 #define mmToIlu(mm) ((mm) / length_quant)
 #define iluToMm(ilu) ((ilu) * length_quant)
 #define FUJITSU_CONFIG_FILE "fujitsu.conf"
+#define READ_MODE_PASS 1
+#define READ_MODE_RRGGBB 2
+#define READ_MODE_3091RGB 3
+#define READ_MODE_BGR 4
+
 
 #define FIXED_MM_TO_SCANNER_UNIT(number) SANE_UNFIX(number) * 1200 / MM_PER_INCH
 #define SCANNER_UNIT_TO_FIXED_MM(number) SANE_FIX(number * MM_PER_INCH / 1200)
@@ -386,7 +440,7 @@ struct fujitsu
 SANE_Status sane_init (SANE_Int * version_code, SANE_Auth_Callback authorize);
 
 SANE_Status sane_get_devices (const SANE_Device *** device_list,
-			      SANE_Bool local_only);
+                              SANE_Bool local_only);
 
 SANE_Status sane_open (SANE_String_Const name, SANE_Handle * handle);
 
@@ -395,19 +449,19 @@ SANE_Status sane_set_io_mode (SANE_Handle h, SANE_Bool non_blocking);
 SANE_Status sane_get_select_fd (SANE_Handle h, SANE_Int * fdp);
 
 const SANE_Option_Descriptor *sane_get_option_descriptor (SANE_Handle handle,
-							  SANE_Int option);
+                                                          SANE_Int option);
 
 SANE_Status sane_control_option (SANE_Handle handle, SANE_Int option,
-				 SANE_Action action, void *val,
-				 SANE_Int * info);
+                                 SANE_Action action, void *val,
+                                 SANE_Int * info);
 
 SANE_Status sane_start (SANE_Handle handle);
 
 SANE_Status sane_get_parameters (SANE_Handle handle,
-				 SANE_Parameters * params);
+                                 SANE_Parameters * params);
 
 SANE_Status sane_read (SANE_Handle handle, SANE_Byte * buf, SANE_Int max_len,
-		       SANE_Int * len);
+                       SANE_Int * len);
 
 void sane_cancel (SANE_Handle h);
 
@@ -417,80 +471,115 @@ void sane_exit (void);
 
 /* ------------------------------------------------------------------------- */
 
-static void doInquiry (struct fujitsu *s);
+static void do_inquiry (struct fujitsu *s);
 
 static SANE_Status attachScanner (const char *devicename,
-				  struct fujitsu **devp);
+                                  struct fujitsu **devp);
 
-static SANE_Status senseHandler (int scsi_fd, u_char * result, void *arg);
+static SANE_Status scsi_sense_handler (int scsi_fd, u_char * result, void *arg);
 
-static int identifyScanner (struct fujitsu *s);
+static int identify_scanner (struct fujitsu *s);
 
-static void doInquiry (struct fujitsu *s);
+static int
+do_cmd (Fujitsu_Connection_Type connection, int fd, unsigned char *cmd,
+             int cmd_len, unsigned char *out, size_t req_out_len,
+             size_t *res_out_len);
 
 static int do_scsi_cmd (int fd, unsigned char *cmd, int cmd_len,
-			unsigned char *out, size_t out_len);
+                        unsigned char *out, size_t req_out_len, 
+                        size_t *res_out_len);
+
+static int
+do_usb_cmd (int fd, unsigned char *cmd,
+             int cmd_len, unsigned char *out, size_t req_out_len,
+             size_t *res_out_len);
 
 static void hexdump (int level, char *comment, unsigned char *p, int l);
 
-static SANE_Status initOptions (struct fujitsu *scanner);
+static SANE_Status init_options (struct fujitsu *scanner);
 
-static int grabScanner (struct fujitsu *s);
+static int grab_scanner (struct fujitsu *s);
 
-static int freeScanner (struct fujitsu *s);
+static int free_scanner (struct fujitsu *s);
 
-static int waitScanner (struct fujitsu *s);
+static int wait_scanner (struct fujitsu *s);
 
-static int objectPosition (struct fujitsu *s);
+static int object_position (struct fujitsu *s, int i_load);
 
-static SANE_Status doCancel (struct fujitsu *scanner);
+static SANE_Status do_cancel (struct fujitsu *scanner);
 
-static int objectDischarge (struct fujitsu *s);
+static SANE_Status do_reset (struct fujitsu *scanner);
+
+/*static int objectDischarge (struct fujitsu *s);*/
+
+static int fujitsu_set_sleep_mode(struct fujitsu *s);
 
 static int set_mode_params (struct fujitsu *s);
 
+static int imprinter(struct fujitsu *s);
+
+static int fujitsu_send(struct fujitsu *s);
+
 static int setWindowParam (struct fujitsu *s);
 
 static size_t maxStringSize (const SANE_String_Const strings[]);
 
-static int startScan (struct fujitsu *s);
+static int start_scan (struct fujitsu *s);
 
 static int reader_process (struct fujitsu *scanner, int fd1, int fd2);
 
 static unsigned int reader3091ColorDuplex (struct fujitsu *scanner, FILE * fd,
-					   FILE * fd2);
+                                           FILE * fd2);
 static unsigned int reader3091ColorSimplex (struct fujitsu *scanner,
-					    FILE * fd);
+                                            FILE * fd);
 static unsigned int reader3091GrayDuplex (struct fujitsu *scanner, FILE * fd,
-					  FILE * fd2);
-static unsigned int reader_gray_duplex_sequential (struct fujitsu *scanner,
-						   FILE * fd, FILE * fd2);
-static unsigned int reader_gray_duplex_alternate (struct fujitsu *scanner,
-						  FILE * fd, FILE * fd2);
+                                          FILE * fd2);
+static unsigned int reader3092ColorDuplex (struct fujitsu *scanner, FILE * fd,
+                                          FILE * fd2);
+static void convert_bgr_to_rgb(struct fujitsu *scanner, unsigned char * buffptr, unsigned int length);
 
-static unsigned int readerGenericPassthrough (struct fujitsu *scanner,
-					      FILE * fd, int i_window_id);
+static void convert_rrggbb_to_rgb(struct fujitsu *scanner, unsigned char * buffptr, unsigned int length);
+
+static unsigned int reader3092ColorSimplex (struct fujitsu *scanner,
+                                           FILE * fd);
+static unsigned int reader3092GrayDuplex (struct fujitsu *scanner, FILE * fd,
+                                         FILE * fd2);
+static unsigned int reader_gray_duplex_sequential (struct fujitsu *scanner,
+                                                   FILE * fd, FILE * fd2);
+static unsigned int reader_gray_duplex_alternate (struct fujitsu *scanner,
+                                                  FILE * fd, FILE * fd2);
+
+static unsigned int reader_duplex_sequential (struct fujitsu *scanner,
+                                                   FILE * fd, FILE * fd2);
+static unsigned int reader_duplex_alternate (struct fujitsu *scanner,
+                                                  FILE * fd, FILE * fd2);
+static unsigned int reader_simplex (struct fujitsu *scanner, FILE * fd, 
+					int i_window_id);
+
+static unsigned int reader_generic_passthrough (struct fujitsu *scanner,
+                                                FILE * fd, int i_window_id);
 
 static int read_large_data_block (struct fujitsu *s,
-				  unsigned char *buffer, unsigned int length,
-				  int i_window_id);
+                                  unsigned char *buffer, unsigned int length,
+                                  int i_window_id, 
+                                  unsigned int *i_data_read);
 
 static SANE_Status attachOne (const char *name);
 
 static int modelMatch (const char *product);
 
 static void setDefaults3091 (struct fujitsu *scanner);
+static void setDefaults3092 (struct fujitsu *scanner);
 static void setDefaults3096 (struct fujitsu *scanner);
-static void setDefaultsSP15 (struct fujitsu *scanner);
 
 static SANE_Status setMode3091 (struct fujitsu *scanner, int mode);
+static SANE_Status setMode3092 (struct fujitsu *scanner, int mode);
 static SANE_Status setMode3096 (struct fujitsu *scanner, int mode);
-static SANE_Status setModeSP15 (struct fujitsu *scanner, int mode);
 
 static void calculateDerivedValues (struct fujitsu *scanner);
 
 static int makeTempFile (void);
-static SANE_Status getHardwareStatus (struct fujitsu *s);
+static SANE_Status get_hardware_status (struct fujitsu *s);
 static void fujitsu_set_standard_size (SANE_Handle handle);
 
 
