@@ -43,9 +43,27 @@
    HP Scanner Control Language (SCL).
 */
 
-static char *hp_backend_version = "1.03";
+static char *hp_backend_version = "1.05";
+static char *hp_backend_revision = "$Revision: 1.19 $";
 /* Changes:
 
+   V 1.05:
+   $Log: hp.c,v $
+   Revision 1.19  2003/10/24 17:26:07  kig-guest
+   Use new sanei-thread-interface
+
+   Revision 1.18  2003/10/09 19:37:29  kig-guest
+   Redo when TEST UNIT READY failed
+   Redo when read returns with 0 bytes (non-SCSI only)
+   Bug #300241: fix invers image on 3c/4c/6100C at 10 bit depth
+
+   Revision 1.17  2003/10/06 19:54:07  kig-guest
+   Bug #300248: correct "Negatives" to "Negative" in option description
+ 
+
+   V 1.04, 24-Jul-2003, PK (peter@kirchgessner.net)
+      - Add internationalization
+
    V 1.03, 14-Apr-2003, PK (peter@kirchgessner.net)
       - check valp in call of sane_control_option()
 
@@ -209,6 +227,7 @@ static char *hp_backend_version = "1.03";
 #include "sane/sanei_config.h"
 #include "sane/sanei_backend.h"
 #include "sane/sanei_usb.h"
+#include "sane/sanei_thread.h"
 /* #include <sane/sanei_debug.h> */
 #include "hp-device.h"
 #include "hp-handle.h"
@@ -623,8 +642,8 @@ hp_read_config (void)
   is_df_config = 1;
   cu_device[0] = '\0';
 
-  DBG(1, "hp_read_config: hp backend v%s starts reading config file\n",
-      hp_backend_version);
+  DBG(1, "hp_read_config: hp backend v%s/%s starts reading config file\n",
+      hp_backend_version, hp_backend_revision);
 
   if ((fp = sanei_config_open(HP_CONFIG_FILE)) != 0)
     {
@@ -783,6 +802,7 @@ sane_init (SANE_Int *version_code, SANE_Auth_Callback UNUSEDARG authorize)
 
   DBG_INIT();
   DBG(3, "sane_init called\n");
+  sanei_thread_init ();
 
   hp_destroy();
 
