@@ -46,9 +46,62 @@
    This file is a component of the implementation of a backend for many
    of the the AGFA SnapScan and Acer Vuego/Prisa flatbed scanners. */
 
-/* $Id: snapscan-options.c,v 1.4 2002/05/02 17:19:14 oliverschwartz Exp $
+/* $Id: snapscan-options.c,v 1.12 2003/11/09 21:43:45 oliver-guest Exp $
    SANE SnapScan backend */
 
+/* default option values */
+
+#define DEFAULT_RES             300
+#define DEFAULT_PREVIEW         SANE_FALSE
+#define DEFAULT_HIGHQUALITY     SANE_FALSE
+#define DEFAULT_BRIGHTNESS      0
+#define DEFAULT_CONTRAST        0
+#define DEFAULT_GAMMA           SANE_FIX(1.8)
+#define DEFAULT_HALFTONE        SANE_FALSE
+#define DEFAULT_NEGATIVE        SANE_FALSE
+#define DEFAULT_THRESHOLD       50
+#define DEFAULT_QUALITY         SANE_TRUE
+#define DEFAULT_CUSTOM_GAMMA    SANE_FALSE
+#define DEFAULT_GAMMA_BIND      SANE_FALSE
+
+static SANE_Int def_rgb_lpr = 4;
+static SANE_Int def_gs_lpr = 12;
+
+
+/* predefined preview mode name */
+static char md_auto[] = "Auto";
+
+/* predefined scan mode names */
+static char md_colour[] = SANE_I18N("Color");
+static char md_bilevelcolour[] = SANE_I18N("Halftone");
+static char md_greyscale[] = SANE_I18N("Gray");
+static char md_lineart[] = SANE_I18N("Lineart");
+
+/* predefined scan source names */
+static char src_flatbed[] = SANE_I18N("Flatbed");
+static char src_tpo[] = SANE_I18N("Transparency Adapter");
+static char src_adf[] = SANE_I18N("Document Feeder");
+
+/* predefined scan window setting names */
+static char pdw_none[] = SANE_I18N("None");
+static char pdw_6X4[] = SANE_I18N("6x4 (inch)");
+static char pdw_8X10[] = SANE_I18N("8x10 (inch)");
+static char pdw_85X11[] = SANE_I18N("8.5x11 (inch)");
+
+/* predefined dither matrix names */
+static char dm_none[] = SANE_I18N("Halftoning Unsupported");
+static char dm_dd8x8[] = SANE_I18N("DispersedDot8x8");
+static char dm_dd16x16[] = SANE_I18N("DispersedDot16x16");
+
+/* strings */
+static char lpr_desc[] = SANE_I18N(
+    "Number of scan lines to request in a SCSI read. "
+    "Changing this parameter allows you to tune the speed at which "
+    "data is read from the scanner during scans. If this is set too "
+    "low, the scanner will have to stop periodically in the middle of "
+    "a scan; if it's set too high, X-based frontends may stop responding "
+    "to X events and your system could bog down.");
+
 /* ranges */
 static const SANE_Range x_range_fb =
 {
@@ -58,6 +111,10 @@ static const SANE_Range y_range_fb =
 {
     SANE_FIX (0.0), SANE_FIX (297.0), 0
 };        /* mm */
+
+/* default TPO range (shortest y_range
+   to avoid tray collision.
+*/
 static const SANE_Range x_range_tpo_default =
 {
     SANE_FIX (0.0), SANE_FIX (129.0), 0
@@ -66,6 +123,8 @@ static const SANE_Range y_range_tpo_default =
 {
     SANE_FIX (0.0), SANE_FIX (180.0), 0
 };        /* mm */
+
+/* TPO range for the Agfa 1236 */
 static const SANE_Range x_range_tpo_1236 =
 {
     SANE_FIX (0.0), SANE_FIX (203.0), 0
@@ -74,6 +133,27 @@ static const SANE_Range y_range_tpo_1236 =
 {
     SANE_FIX (0.0), SANE_FIX (254.0), 0
 };        /* mm */
+
+/* TPO range for the Agfa e50 */
+static const SANE_Range x_range_tpo_e50 =
+{
+    SANE_FIX (0.0), SANE_FIX (40.0), 0
+};        /* mm */
+static const SANE_Range y_range_tpo_e50 =
+{
+    SANE_FIX (0.0), SANE_FIX (240.0), 0
+};        /* mm */
+
+/* TPO range for the Epson 1670 */
+static const SANE_Range x_range_tpo_1670 =
+{
+    SANE_FIX (0.0), SANE_FIX (101.0), 0
+};        /* mm */
+static const SANE_Range y_range_tpo_1670 =
+{
+    SANE_FIX (0.0), SANE_FIX (228.0), 0
+};        /* mm */
+
 static SANE_Range x_range_tpo;
 static SANE_Range y_range_tpo;
 static const SANE_Range gamma_range =
@@ -125,6 +205,8 @@ static void init_options (SnapScan_Scanner * ps)
         {8, 50, 75, 100, 150, 200, 300, 450, 600};
     static SANE_Word resolutions_1200[] =
         {10, 50, 75, 100, 150, 200, 300, 450, 600, 900, 1200};
+    static SANE_Word resolutions_1600[] =
+        {10, 50, 75, 100, 150, 200, 300, 450, 600, 900, 1200, 1600};
     static SANE_String_Const names_all[] =
         {md_colour, md_bilevelcolour, md_greyscale, md_lineart, NULL};
     static SANE_String_Const names_basic[] =
@@ -135,6 +217,30 @@ static void init_options (SnapScan_Scanner * ps)
         {md_auto, md_colour, md_greyscale, md_lineart, NULL};
     SANE_Option_Descriptor *po = ps->options;
 
+    /* Initialize TPO range */
+    switch (ps->pdev->model)
+    {
+    case SNAPSCAN1236:
+        x_range_tpo = x_range_tpo_1236;
+        y_range_tpo = y_range_tpo_1236;
+        break;
+    case SNAPSCANE20:
+    case SNAPSCANE50:
+    case SNAPSCANE52:
+        x_range_tpo = x_range_tpo_e50;
+        y_range_tpo = y_range_tpo_e50;
+        break;
+    case PERFECTION1670:
+        x_range_tpo = x_range_tpo_1670;
+        y_range_tpo = y_range_tpo_1670;
+        break;
+    default:
+        x_range_tpo = x_range_tpo_default;
+        y_range_tpo = y_range_tpo_default;
+        break;
+    }
+
+	/* Initialize option descriptors */
     po[OPT_COUNT].name = SANE_NAME_NUM_OPTIONS;
     po[OPT_COUNT].title = SANE_TITLE_NUM_OPTIONS;
     po[OPT_COUNT].desc = SANE_DESC_NUM_OPTIONS;
@@ -167,15 +273,20 @@ static void init_options (SnapScan_Scanner * ps)
     switch (ps->pdev->model)
     {
     case SNAPSCAN310:
-    case VUEGO310S:                /* WG changed */
+    case PRISA310:                /* WG changed */
         po[OPT_SCANRES].constraint.word_list = resolutions_300;
         break;
     case SNAPSCANE50:
     case SNAPSCANE52:
     case PRISA5300:
     case PRISA1240:
+    case PRISA5000:
+    case ARCUS1200:
         po[OPT_SCANRES].constraint.word_list = resolutions_1200;
         break;
+    case PERFECTION1670:
+        po[OPT_SCANRES].constraint.word_list = resolutions_1600;
+        break;
     default:
         po[OPT_SCANRES].constraint.word_list = resolutions_600;
         break;
@@ -193,6 +304,17 @@ static void init_options (SnapScan_Scanner * ps)
     po[OPT_PREVIEW].constraint_type = SANE_CONSTRAINT_NONE;
     ps->preview = DEFAULT_PREVIEW;
 
+    po[OPT_HIGHQUALITY].name = "high-quality";
+    po[OPT_HIGHQUALITY].title = SANE_I18N("High quality");
+    po[OPT_HIGHQUALITY].desc = SANE_I18N("Scan in (slower) high quality mode");
+    po[OPT_HIGHQUALITY].type = SANE_TYPE_BOOL;
+    po[OPT_HIGHQUALITY].unit = SANE_UNIT_NONE;
+    po[OPT_HIGHQUALITY].size = sizeof (SANE_Word);
+    po[OPT_HIGHQUALITY].cap =
+        SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT | SANE_CAP_AUTOMATIC;
+    po[OPT_HIGHQUALITY].constraint_type = SANE_CONSTRAINT_NONE;
+    ps->highquality = DEFAULT_HIGHQUALITY;
+
     po[OPT_BRIGHTNESS].name = SANE_NAME_BRIGHTNESS;
     po[OPT_BRIGHTNESS].title = SANE_TITLE_BRIGHTNESS;
     po[OPT_BRIGHTNESS].desc = SANE_DESC_BRIGHTNESS;
@@ -226,7 +348,7 @@ static void init_options (SnapScan_Scanner * ps)
     switch (ps->pdev->model)
     {
     case SNAPSCAN310:
-    case VUEGO310S:
+    case PRISA310:
         po[OPT_MODE].constraint.string_list = names_basic;
         break;
     default:
@@ -244,15 +366,15 @@ static void init_options (SnapScan_Scanner * ps)
     po[OPT_PREVIEW_MODE].type = SANE_TYPE_STRING;
     po[OPT_PREVIEW_MODE].unit = SANE_UNIT_NONE;
     po[OPT_PREVIEW_MODE].size = 32;
-    po[OPT_PREVIEW_MODE].cap = SANE_CAP_SOFT_SELECT 
+    po[OPT_PREVIEW_MODE].cap = SANE_CAP_SOFT_SELECT
                                | SANE_CAP_SOFT_DETECT
-                               | SANE_CAP_ADVANCED 
+                               | SANE_CAP_ADVANCED
                                | SANE_CAP_AUTOMATIC;
     po[OPT_PREVIEW_MODE].constraint_type = SANE_CONSTRAINT_STRING_LIST;
     switch (ps->pdev->model)
     {
     case SNAPSCAN310:
-    case VUEGO310S:
+    case PRISA310:
         po[OPT_PREVIEW_MODE].constraint.string_list = preview_names_basic;
         break;
     default:
@@ -377,10 +499,12 @@ static void init_options (SnapScan_Scanner * ps)
     po[OPT_QUALITY_CAL].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
     ps->val[OPT_QUALITY_CAL].b = DEFAULT_QUALITY;
     /* Disable quality calibration option if not supported
-       Note: Snapscan e52 does not support quality calibration,
+       Note: Snapscan e52 and Prisa5300 do not support quality calibration,
        although HCFG_CAL_ALLOWED is set. */
     if ((!(ps->hconfig & HCFG_CAL_ALLOWED))
-        || (ps->pdev->model == SNAPSCANE52)) {
+        || (ps->pdev->model == SNAPSCANE52)
+        || (ps->pdev->model == PERFECTION1670)
+        || (ps->pdev->model == PRISA5300)) {
         po[OPT_QUALITY_CAL].cap |= SANE_CAP_INACTIVE;
         ps->val[OPT_QUALITY_CAL].b = SANE_FALSE;
     }
@@ -738,6 +862,9 @@ SANE_Status sane_control_option (SANE_Handle h,
         case OPT_PREVIEW:
             *(SANE_Bool *) v = pss->preview;
             break;
+        case OPT_HIGHQUALITY:
+            *(SANE_Bool *) v = pss->highquality;
+            break;
         case OPT_MODE:
             DBG (DL_VERBOSE,
                  "%s: writing \"%s\" to location %p\n",
@@ -847,7 +974,7 @@ SANE_Status sane_control_option (SANE_Handle h,
             return SANE_STATUS_INVAL;
         }
         /* prevent setting of options during a scan */
-        if (pss->state!=ST_IDLE) {
+        if ((pss->state==ST_SCAN_INIT) || (pss->state==ST_SCANNING)) {
             DBG(DL_INFO,
                 "set value for option %s ignored: scanner is still scanning (status %d)\n",
                 pss->options[n].name,
@@ -873,6 +1000,11 @@ SANE_Status sane_control_option (SANE_Handle h,
             if (i)
                 *i |= SANE_INFO_RELOAD_PARAMS;
             break;
+        case OPT_HIGHQUALITY:
+            pss->preview = *(SANE_Bool *) v;
+            if (i)
+                *i |= SANE_INFO_RELOAD_PARAMS;
+            break;
         case OPT_MODE:
             {
                 char *s = (SANE_String) v;
@@ -1274,6 +1406,11 @@ SANE_Status sane_control_option (SANE_Handle h,
             if (i)
                 *i |= SANE_INFO_RELOAD_PARAMS;
             break;
+        case OPT_HIGHQUALITY:
+            pss->highquality = SANE_FALSE;
+            if (i)
+                *i |= SANE_INFO_RELOAD_PARAMS;
+            break;
         case OPT_MODE:
             pss->mode_s = md_colour;
             pss->mode = MD_COLOUR;
@@ -1348,11 +1485,38 @@ SANE_Status sane_control_option (SANE_Handle h,
 
 /*
  * $Log: snapscan-options.c,v $
- * Revision 1.4  2002/05/02 17:19:14  oliverschwartz
- * SnapScan backend 1.4.13: Support for ADF
+ * Revision 1.12  2003/11/09 21:43:45  oliver-guest
+ * Disabled quality calibration for Epson Perfection 1670
  *
- * Revision 1.3  2002/04/27 15:35:16  oliverschwartz
- * SnapScan backend 1.4.12: Fix option handling
+ * Revision 1.11  2003/11/08 09:50:27  oliver-guest
+ * Fix TPO scanning range for Epson 1670
+ *
+ * Revision 1.10  2003/10/21 20:43:25  oliver-guest
+ * Bugfixes for SnapScan backend
+ *
+ * Revision 1.9  2003/10/07 18:29:20  oliver-guest
+ * Initial support for Epson 1670, minor bugfix
+ *
+ * Revision 1.8  2003/08/19 21:05:08  oliverschwartz
+ * Scanner ID cleanup
+ *
+ * Revision 1.7  2003/04/30 20:49:39  oliverschwartz
+ * SnapScan backend 1.4.26
+ *
+ * Revision 1.8  2003/04/30 20:42:18  oliverschwartz
+ * Added support for Agfa Arcus 1200 (supplied by Valtteri Vuorikoski)
+ *
+ * Revision 1.7  2003/04/02 21:17:12  oliverschwartz
+ * Fix for 1200 DPI with Acer 5000
+ *
+ * Revision 1.6  2002/07/12 23:23:06  oliverschwartz
+ * Disable quality calibration for 5300
+ *
+ * Revision 1.5  2002/06/06 20:40:00  oliverschwartz
+ * Changed default scan area for transparancy unit of SnapScan e50
+ *
+ * Revision 1.4  2002/05/02 18:28:44  oliverschwartz
+ * Added ADF support
  *
  * Revision 1.3  2002/04/27 14:43:59  oliverschwartz
  * - Remove SCSI debug options
