@@ -2,7 +2,7 @@
    Uses the SANE library.
    Copyright (C) 1996, 1997, 1998 Andreas Beck and David Mosberger
 
-   You can contact me at becka@sunserver1.rz.uni-duesseldorf.de
+   You can contact me at becka@uni-duesseldorf.de
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -19,7 +19,7 @@
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
 #ifdef _AIX
-# include "../include/lalloca.h"		/* MUST come first for AIX! */
+# include "../include/lalloca.h"	/* MUST come first for AIX! */
 #endif
 
 #include "../include/sane/config.h"
@@ -88,13 +88,15 @@ static struct option basic_options[] = {
   {"batch-increment", required_argument, NULL, OPTION_BATCH_INCREMENT},
   {"format", required_argument, NULL, OPTION_FORMAT},
   {"accept-md5-only", no_argument, NULL, OPTION_MD5},
+  {"icc-profile", required_argument, NULL, 'i'},
+  {"dont-scan", no_argument, NULL, 'n'},
   {0, 0, NULL, 0}
 };
 
 #define OUTPUT_PNM      0
 #define OUTPUT_TIFF     1
 
-#define BASE_OPTSTRING	"d:hLf:vVTb"
+#define BASE_OPTSTRING	"d:hi:Lf:nvVTb"
 #define STRIP_HEIGHT	256	/* # lines we increment image height */
 
 static struct option *all_options;
@@ -105,6 +107,7 @@ static int verbose;
 static int test;
 static int output_format = OUTPUT_PNM;
 static int help;
+static int dont_scan = 0;
 static const char *prog_name;
 static SANE_Option_Descriptor window_option[2];
 static int window[4];
@@ -112,6 +115,7 @@ static int resolution_optind = -1, resolution_value = 0;
 static SANE_Word window_val[2];
 static int window_val_user[2];	/* is width/height user-specified? */
 static int accept_only_md5_auth = 0;
+static const char *icc_profile = NULL;
 
 static void fetch_options (SANE_Device * device);
 static void scanimage_exit (void);
@@ -170,8 +174,7 @@ auth_callback (SANE_String_Const resource,
 		      if (strchr (strchr (tmp, ':') + 1, ':') != NULL)
 			{
 
-			  if (
-			      (strncmp
+			  if ((strncmp
 			       (strchr (strchr (tmp, ':') + 1, ':') + 1,
 				resource, len) == 0)
 			      &&
@@ -184,8 +187,7 @@ auth_callback (SANE_String_Const resource,
 				  SANE_MAX_USERNAME_LEN)
 				{
 
-				  if (
-				      (strchr (strchr (tmp, ':') + 1, ':') -
+				  if ((strchr (strchr (tmp, ':') + 1, ':') -
 				       (strchr (tmp, ':') + 1)) <
 				      SANE_MAX_PASSWORD_LEN)
 				    {
@@ -291,8 +293,7 @@ sighandler (int signum)
 
   if (device)
     {
-      fprintf (stderr, "%s: received signal %d\n", prog_name,
-	       signum);
+      fprintf (stderr, "%s: received signal %d\n", prog_name, signum);
       if (first_time)
 	{
 	  first_time = SANE_FALSE;
@@ -302,7 +303,7 @@ sighandler (int signum)
       else
 	{
 	  fprintf (stderr, "%s: aborting\n", prog_name);
-	  _exit(0);
+	  _exit (0);
 	}
     }
 }
@@ -444,12 +445,12 @@ print_option (SANE_Device * device, int opt_num, char short_name)
   if (opt->type == SANE_TYPE_STRING || opt->size == sizeof (SANE_Word))
     {
       /* print current option value */
-      fputs (" [", stdout);
       if (SANE_OPTION_IS_ACTIVE (opt->cap))
 	{
 	  void *val = alloca (opt->size);
 	  sane_control_option (device, opt_num, SANE_ACTION_GET_VALUE, val,
 			       0);
+	  fputs (" [", stdout);
 	  switch (opt->type)
 	    {
 	    case SANE_TYPE_BOOL:
@@ -471,11 +472,13 @@ print_option (SANE_Device * device, int opt_num, char short_name)
 	    default:
 	      break;
 	    }
+	  fputc (']', stdout);
 	}
-      else
-	fputs ("inactive", stdout);
-      fputc (']', stdout);
     }
+
+  if (!SANE_OPTION_IS_ACTIVE (opt->cap))
+    fputs (" [inactive]", stdout);
+
   fputs ("\n        ", stdout);
 
   switch (short_name)
@@ -876,7 +879,7 @@ process_backend_option (SANE_Handle device, int optnum, const char *optarg)
       exit (1);
     }
 
-  if ((opt->cap & SANE_CAP_AUTOMATIC) && optarg && 
+  if ((opt->cap & SANE_CAP_AUTOMATIC) && optarg &&
       strncasecmp (optarg, "auto", 4) == 0)
     {
       status = sane_control_option (device, optnum, SANE_ACTION_SET_AUTO,
@@ -930,7 +933,14 @@ process_backend_option (SANE_Handle device, int optnum, const char *optarg)
       break;
 
     case SANE_TYPE_STRING:
-      valuep = (void *) optarg;
+      valuep = malloc (opt->size);
+      if (!valuep)
+	{
+	  fprintf (stderr, "%s: out of memory\n", prog_name);
+	  exit (1);
+	}
+      strncpy (valuep, optarg, opt->size);
+      ((char *) valuep)[opt->size - 1] = 0;
       break;
 
     case SANE_TYPE_BUTTON:
@@ -1082,7 +1092,7 @@ scan_it (void)
 		  if (output_format == OUTPUT_TIFF)
 		    sanei_write_tiff_header (parm.format,
 					     parm.pixels_per_line, parm.lines,
-					     parm.depth, resolution_value);
+					     parm.depth, resolution_value, icc_profile);
 		  else
 		    write_pnm_header (parm.format, parm.pixels_per_line,
 				      parm.lines, parm.depth);
@@ -1192,17 +1202,17 @@ scan_it (void)
 		  break;
 		}
 	    }
-	  else /* ! must_buffer */
-  	    {
+	  else			/* ! must_buffer */
+	    {
 	      if ((output_format == OUTPUT_TIFF) || (parm.depth != 16))
 		fwrite (buffer, 1, len, stdout);
 	      else
-	        {
+		{
 #if !defined(WORDS_BIGENDIAN)
 		  int i, start = 0;
-		  
+
 		  /* check if we have saved one byte from the last sane_read */
-		  if (hang_over > -1) 
+		  if (hang_over > -1)
 		    {
 		      if (len > 0)
 			{
@@ -1217,13 +1227,13 @@ scan_it (void)
 		    {
 		      unsigned char LSB;
 		      LSB = buffer[i];
-		      buffer[i] = buffer[i+1];
-		      buffer[i+1] = LSB;
+		      buffer[i] = buffer[i + 1];
+		      buffer[i + 1] = LSB;
 		    }
 		  /* check if we have an odd number of bytes */
 		  if (((len - start) % 2) != 0)
 		    {
-		      hang_over = buffer [len - 1];
+		      hang_over = buffer[len - 1];
 		      len--;
 		    }
 #endif
@@ -1249,22 +1259,22 @@ scan_it (void)
       image.height = image.y;
       if (output_format == OUTPUT_TIFF)
 	sanei_write_tiff_header (parm.format, parm.pixels_per_line,
-				 parm.lines, parm.depth, resolution_value);
+				 parm.lines, parm.depth, resolution_value, icc_profile);
       else
 	write_pnm_header (parm.format, image.width, image.height, parm.depth);
-      if ((output_format == OUTPUT_TIFF) || (image.Bpp == 1) 
+      if ((output_format == OUTPUT_TIFF) || (image.Bpp == 1)
 	  || (image.Bpp == 3))
 	fwrite (image.data, image.Bpp, image.height * image.width, stdout);
-      else /* image.Bpp == 2 or image.Bpp == 6 assumed */
-        {
+      else			/* image.Bpp == 2 or image.Bpp == 6 assumed */
+	{
 #if !defined(WORDS_BIGENDIAN)
 	  int i;
 	  for (i = 0; i < image.Bpp * image.height * image.width; i += 2)
 	    {
 	      unsigned char LSB;
 	      LSB = image.data[i];
-	      image.data[i] = image.data[i+1];
-	      image.data[i+1] = LSB;
+	      image.data[i] = image.data[i + 1];
+	      image.data[i + 1] = LSB;
 	    }
 #endif
 	  fwrite (image.data, image.Bpp, image.height * image.width, stdout);
@@ -1277,8 +1287,9 @@ cleanup:
     free (image.data);
 
 
-  expected_bytes = parm.bytes_per_line * parm.lines * 
-    ((parm.format == SANE_FRAME_RGB || parm.format == SANE_FRAME_GRAY)?1:3);
+  expected_bytes = parm.bytes_per_line * parm.lines *
+    ((parm.format == SANE_FRAME_RGB
+      || parm.format == SANE_FRAME_GRAY) ? 1 : 3);
   if (parm.lines < 0)
     expected_bytes = 0;
   if (total_bytes > expected_bytes && expected_bytes != 0)
@@ -1429,11 +1440,11 @@ scanimage_exit (void)
 {
   if (device)
     {
-      if (verbose > 1) 
+      if (verbose > 1)
 	fprintf (stderr, "Closing device\n");
       sane_close (device);
     }
-  if (verbose > 1) 
+  if (verbose > 1)
     fprintf (stderr, "Calling sane_exit\n");
   sane_exit ();
 
@@ -1441,7 +1452,7 @@ scanimage_exit (void)
     free (all_options);
   if (option_number)
     free (option_number);
-  if (verbose > 1) 
+  if (verbose > 1)
     fprintf (stderr, "scanimage: finished\n");
 }
 
@@ -1486,7 +1497,6 @@ main (int argc, char **argv)
 	case ':':
 	case '?':
 	  break;		/* may be an option that we'll parse later on */
-
 	case 'd':
 	  devname = optarg;
 	  break;
@@ -1498,23 +1508,29 @@ main (int argc, char **argv)
 	case 'h':
 	  help = 1;
 	  break;
+        case 'i': /* icc profile */
+          icc_profile = optarg;
+          break;
 	case 'v':
 	  ++verbose;
 	  break;
 	case 'T':
 	  test = 1;
 	  break;
+	case 'n':
+	  dont_scan = 1;
+	  break;
 	case OPTION_BATCH_INCREMENT:
-	  batch_increment = atoi(optarg);
+	  batch_increment = atoi (optarg);
 	  break;
 	case OPTION_BATCH_START_AT:
-	  batch_start_at = atoi(optarg);
+	  batch_start_at = atoi (optarg);
 	  break;
 	case OPTION_BATCH_DOUBLE:
 	  batch_increment = 2;
 	  break;
 	case OPTION_BATCH_COUNT:
-	  batch_count = atoi(optarg);
+	  batch_count = atoi (optarg);
 	  batch = 1;
 	  break;
 	case OPTION_FORMAT:
@@ -1539,97 +1555,98 @@ main (int argc, char **argv)
 		exit (1);
 	      }
 
-            if (ch == 'L')
-              {
-	        for (i = 0; device_list[i]; ++i)
-	          {
+	    if (ch == 'L')
+	      {
+		for (i = 0; device_list[i]; ++i)
+		  {
 		    printf ("device `%s' is a %s %s %s\n",
 			    device_list[i]->name, device_list[i]->vendor,
 			    device_list[i]->model, device_list[i]->type);
-	          }
+		  }
 	      }
-	    else 
+	    else
 	      {
-	        int i = 0, int_arg = 0;
-	        char *percent, *start, *fmt;
-	        const char *text_arg = 0;
-	        char cc, ftype;
-	        
-	        fmt = malloc(strlen(optarg)+1);
-	        if (fmt == 0)
-	          {
-	            fprintf(stderr, "%s: not enough memory\n", prog_name);
-	            exit(1);
-	          }
-	        
-	        for (i = 0; device_list[i]; ++i)
-	          {
-	            strcpy(fmt, optarg);
-	            start = fmt;
-	            while(*start && (percent = strchr(start, '%')) )
-	              {
-	                percent++;
-  	                if (*percent)
-	                  {
-	                    switch (*percent)
-	                      {
-	                        case 'd':
-	                          text_arg = device_list[i]->name;
-	                          ftype = *percent = 's';
-	                          break;
-	                        case 'v':
-	                          text_arg = device_list[i]->vendor;
-	                          ftype = *percent = 's';
-	                          break;
-	                        case 'm':
-	                          text_arg = device_list[i]->model;
-	                          ftype = *percent = 's';
-	                          break;
-	                        case 't':
-	                          text_arg = device_list[i]->type;
-	                          ftype = *percent = 's';
-	                          break;
-	                        case 'i':
-	                          int_arg = i;
-	                          ftype = 'i';
-	                          break;
-	                        case '%':
-	                          ftype = 0;
-	                          break;
-	                        default:
-	                          fprintf(stderr, "%s: unknown format specifier %%%c\n",
-					  prog_name, *percent);
-				  *percent = '%';
-				  ftype = 0;
-	                      }
-	                    percent++;
-	                    cc = *percent;
-	                    *percent = 0;
-	                    switch (ftype)
-	                      {
-	                        case 's':
-	                          printf(start, text_arg);
-	                          break;
-	                        case 'i':
-	                          printf(start, int_arg);
-	                          break;
-	                        case 0:
-	                          printf(start);
-	                          break;
-	                      }
-	                    *percent = cc;
-	                    start = percent;
-	                  }
-	                else
-	                  {
-	                    /* last char of the string is a '%', suppress it */
-	                    *start = 0;
-	                    break;
-	                  }
-	              }
-	            if (*start)
-	              printf(start);
-	          }
+		int i = 0, int_arg = 0;
+		char *percent, *start, *fmt;
+		const char *text_arg = 0;
+		char cc, ftype;
+
+		fmt = malloc (strlen (optarg) + 1);
+		if (fmt == 0)
+		  {
+		    fprintf (stderr, "%s: not enough memory\n", prog_name);
+		    exit (1);
+		  }
+
+		for (i = 0; device_list[i]; ++i)
+		  {
+		    strcpy (fmt, optarg);
+		    start = fmt;
+		    while (*start && (percent = strchr (start, '%')))
+		      {
+			percent++;
+			if (*percent)
+			  {
+			    switch (*percent)
+			      {
+			      case 'd':
+				text_arg = device_list[i]->name;
+				ftype = *percent = 's';
+				break;
+			      case 'v':
+				text_arg = device_list[i]->vendor;
+				ftype = *percent = 's';
+				break;
+			      case 'm':
+				text_arg = device_list[i]->model;
+				ftype = *percent = 's';
+				break;
+			      case 't':
+				text_arg = device_list[i]->type;
+				ftype = *percent = 's';
+				break;
+			      case 'i':
+				int_arg = i;
+				ftype = 'i';
+				break;
+			      case '%':
+				ftype = 0;
+				break;
+			      default:
+				fprintf (stderr,
+					 "%s: unknown format specifier %%%c\n",
+					 prog_name, *percent);
+				*percent = '%';
+				ftype = 0;
+			      }
+			    percent++;
+			    cc = *percent;
+			    *percent = 0;
+			    switch (ftype)
+			      {
+			      case 's':
+				printf (start, text_arg);
+				break;
+			      case 'i':
+				printf (start, int_arg);
+				break;
+			      case 0:
+				printf (start);
+				break;
+			      }
+			    *percent = cc;
+			    start = percent;
+			  }
+			else
+			  {
+			    /* last char of the string is a '%', suppress it */
+			    *start = 0;
+			    break;
+			  }
+		      }
+		    if (*start)
+		      printf (start);
+		  }
 	      }
 	    if (i == 0 && ch != 'f')
 	      printf ("\nNo scanners were identified. If you were expecting "
@@ -1645,9 +1662,9 @@ main (int argc, char **argv)
 
 	case 'V':
 	  printf ("scanimage (%s) %s; backend version %d.%d.%d\n", PACKAGE,
-		  VERSION, SANE_VERSION_MAJOR(version_code),
-		  SANE_VERSION_MINOR(version_code),
-		  SANE_VERSION_BUILD(version_code));
+		  VERSION, SANE_VERSION_MAJOR (version_code),
+		  SANE_VERSION_MINOR (version_code),
+		  SANE_VERSION_BUILD (version_code));
 	  exit (0);
 
 	default:
@@ -1661,6 +1678,13 @@ main (int argc, char **argv)
 Start image acquisition on a scanner device and write PNM image data to\n\
 standard output.\n\
 \n\
+-d, --device-name=DEVICE   use a given scanner device (e.g. hp:/dev/scanner)\n\
+    --format=pnm|tiff      file format of output file\n\
+-i, --icc-profile=PROFILE  include this ICC profile into TIFF file\n\
+-L, --list-devices         show available scanner devices\n\
+-f, --formatted-device-list=FORMAT similar to -L, but the FORMAT of the output\n\
+                           can be specified: %%d (device name), %%v (vendor),\n\
+                           %%m (model), %%t (type), and %%i (index number)\n\
 -b, --batch[=FORMAT]       working in batch mode, FORMAT is `out%%d.pnm' or\n\
                            `out%%d.tif' by default depending on --format\n\
     --batch-start=#        page number to start naming files with\n\
@@ -1668,18 +1692,12 @@ standard output.\n\
     --batch-increment=#    increase number in filename by an amount of #\n\
     --batch-double         increment page number by two for 2sided originals\n\
                            being scanned in a single sided scanner\n\
-    --format=pnm|tiff      file format of output file\n\
--d, --device-name=DEVICE   use a given scanner device (e.g. hp:/dev/scanner)\n\
--h, --help                 display this help message and exit\n\
--L, --list-devices         show available scanner devices\n\
--f, --formatted-device-list=FORMAT similar to -L, but the FORMAT of the output\n\
-                           can be specified: %%d (device name), %%v (vendor),\n\
-                           %%m (model), %%t (type), and %%i (index number)\n\
+    --accept-md5-only      only accept authorization requests using md5\n\
+-n, --dont-scan            only set options, don't actually scan\n\
 -T, --test                 test backend thoroughly\n\
+-h, --help                 display this help message and exit\n\
 -v, --verbose              give even more status messages\n\
--V, --version              print version information\n\
-    --accept-md5-only      only accept authorization requests using md5\n",
-	    prog_name);
+-V, --version              print version information\n", prog_name);
 
   if (!devname)
     {
@@ -1785,7 +1803,7 @@ standard output.\n\
       }
 
       optind = 0;
-      opterr = 1;	      /* re-enable error printing and arg permutation */
+      opterr = 1;		/* re-enable error printing and arg permutation */
       while ((ch = getopt_long (argc, argv, full_optstring, all_options,
 				&index)) != EOF)
 	{
@@ -1793,7 +1811,7 @@ standard output.\n\
 	    {
 	    case ':':
 	    case '?':
-	      exit (1);	      /* error message is printed by getopt_long() */
+	      exit (1);		/* error message is printed by getopt_long() */
 
 	    case 'd':
 	    case 'h':
@@ -1826,6 +1844,13 @@ standard output.\n\
 	      break;
 	    }
 	}
+      if (optind < argc)
+	{
+	  fprintf (stderr, "%s: argument without option: `%s'; ", prog_name,
+		   argv[argc - 1]);
+	  fprintf (stderr, "try %s --help\n", prog_name);
+	  exit (1);
+	}
 
       free (full_optstring);
       for (index = 0; index < 2; ++index)
@@ -1903,6 +1928,9 @@ List of available devices:", prog_name);
       exit (0);
     }
 
+  if (dont_scan)
+    exit (0);
+
   if (output_format != OUTPUT_PNM)
     resolution_value = get_resolution ();
 
@@ -1923,10 +1951,10 @@ List of available devices:", prog_name);
 	    format = "out%d.pnm";
 	}
 
-      if(batch) {
-        fprintf(stderr, "Scanning %d pages, incrementing by %d, numbering from %d\n",
-	    batch_count, batch_increment, batch_start_at);
-      }
+      if (batch)
+	fprintf (stderr,
+		 "Scanning %d pages, incrementing by %d, numbering from %d\n",
+		 batch_count, batch_increment, batch_start_at);
 
       do
 	{
@@ -1940,14 +1968,16 @@ List of available devices:", prog_name);
 	      return SANE_STATUS_ACCESS_DENIED;
 	    }
 
-	  if (batch) {
-	    fprintf(stderr, "Scanning page %d\n", n);
-	  }
+	  if (batch)
+	    {
+	      fprintf (stderr, "Scanning page %d\n", n);
+	    }
 	  status = scan_it ();
-	  if (batch) {
-	    fprintf (stderr, "Scanned page %d.", n);
-	    fprintf (stderr, " (scanner status = %d)\n", status);
-	  }
+	  if (batch)
+	    {
+	      fprintf (stderr, "Scanned page %d.", n);
+	      fprintf (stderr, " (scanner status = %d)\n", status);
+	    }
 
 	  switch (status)
 	    {
@@ -1966,8 +1996,9 @@ List of available devices:", prog_name);
 	    }			/* switch */
 	  n += batch_increment;
 	}
-      while ((batch && (batch_count==BATCH_COUNT_UNLIMITED || --batch_count))
-	      && SANE_STATUS_GOOD == status);
+      while ((batch
+	      && (batch_count == BATCH_COUNT_UNLIMITED || --batch_count))
+	     && SANE_STATUS_GOOD == status);
     }
   else
     status = test_it ();
